/*  pdnmesh - a 2D finite element solver
    Copyright (C) 2001-2005 Sarod Yatawatta <sarod@users.sf.net>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  $Id: eig_arpack.c,v 1.7 2005/04/21 23:16:54 sarod Exp $
*/
/* ARPACK Solver */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_PTHREAD
#include <pthread.h>
#endif
#include "types.h"

int arpack_max_iterations=300;
int arpack_subspace_dimension=20;

/* prototype for the fortran function */
extern void ardrive_(int *N, int *Nz, double *AA, int *JA, int *IA,
                int *nev, int *ncv, int *lworkl, double *resid,
								double *v, double *workd, double *workl, int *info, int *ido,
								double *select_, double *d, int *ierr,
                 int *NL, double *LL, int *JL, int *IL,
                int *NU, double *UU, int *JU, int *IU, int *maxit);

#ifdef HAVE_PTHREAD
typedef struct t_barrier_ {
  int tcount; /* current no. of threads inside barrier */
  int nthreads; /* the no. of threads the barrier works
                with. This is a constant */
  pthread_mutex_t enter_mutex;
  pthread_mutex_t exit_mutex;
  pthread_cond_t lastthread_cond;
  pthread_cond_t exit_cond;
} th_barrier;

/* initialize barrier */
/* N - no. of accomodated threads */
static void
init_th_barrier(th_barrier *barrier, int N)
{
 barrier->tcount=0; /* initially empty */
 barrier->nthreads=N;
 pthread_mutex_init(&barrier->enter_mutex,NULL);
 pthread_mutex_init(&barrier->exit_mutex,NULL);
 pthread_cond_init(&barrier->lastthread_cond,NULL);
 pthread_cond_init(&barrier->exit_cond,NULL);
}
/* destroy barrier */
static void
destroy_th_barrier(th_barrier *barrier)
{
 pthread_mutex_destroy(&barrier->enter_mutex);
 pthread_mutex_destroy(&barrier->exit_mutex);
 pthread_cond_destroy(&barrier->lastthread_cond);
 pthread_cond_destroy(&barrier->exit_cond);
 barrier->tcount=barrier->nthreads=0;
}
/* the main operation of the barrier */
static void
sync_barrier(th_barrier *barrier)
{
 /* trivial case */
 if(barrier->nthreads <2) return;
 /* else */
 /* new threads enters the barrier. Now close the entry door
  so that other threads cannot enter the barrier until we are done */
 pthread_mutex_lock(&barrier->enter_mutex);
 /* next lock the exit mutex - no threads can leave either */
 pthread_mutex_lock(&barrier->exit_mutex);
 /* now check to see if this is the last expected thread */
 if( ++(barrier->tcount) < barrier->nthreads) {
  /* no. this is not the last thread. so open the entry door */
  pthread_mutex_unlock(&barrier->enter_mutex);
 /* go to sleep */
  pthread_cond_wait(&barrier->exit_cond,&barrier->exit_mutex);
 } else {
 /* this is the last thread */
 /* wakeup sleeping threads */
 pthread_cond_broadcast(&barrier->exit_cond);
 /* go to sleep until all threads are woken up
   and leave the barrier */
 pthread_cond_wait(&barrier->lastthread_cond,&barrier->exit_mutex);
/* now all threads have left the barrier. so open the entry door again */
 pthread_mutex_unlock(&barrier->enter_mutex);
 } 
 /* next to the last thread leaving the barrier */
 if(--(barrier->tcount)==1) {
  /* wakeup the last sleeping thread */
  pthread_cond_broadcast(&barrier->lastthread_cond);
 }
 pthread_mutex_unlock(&barrier->exit_mutex);
} 

/* structure to pass to the LU threads */ 
typedef struct tdata_ {
  th_barrier *barrier;
  int myid;
  int *N;
  SPMat *A;
  SPMat *L;
  int *T; /* no. of threads */
  pthread_mutex_t *lock_mutex; /* lock for updating L*/
} tdata;

static void *
work_threadfn(void *data)
{ 
  tdata *t=(tdata*)data;
  int c; /* current column */
  int i,k;
  double temp,ll,aa,bb;
  int N=SPM_dim(t->A);

  for (c=0; c<N; c++) {
#ifdef DEBUG
   printf("col %d, enter id=%d\n",c,t->myid);
#endif
   /* the 0-th thread will calculate the diagonal entry */
    if(t->myid==0) {
   temp=0;
   /* diagonal element of row j */
   for (i=0;i<c; i++) {
    aa=SPM_e(t->L,c,i);
    if(aa)
     temp+=aa*aa;
   }
   temp=SPM_e(t->A,c,c)-temp;

   if(temp<=0) {
    fprintf(stderr,"%s: %d: matrix not positive definite\n",__FILE__,__LINE__);
    temp=TOL;
   } else {
    temp=sqrt(temp);
   }
   pthread_mutex_lock(t->lock_mutex);
   SPM_eq(t->L,c,c,temp);
   pthread_mutex_unlock(t->lock_mutex);
#ifdef DEBUG
    printf("thread %d, L[%d][%d]=%lf\n",t->myid,c,c,temp);
#endif
    }
   
   sync_barrier(t->barrier);
   /* other threads will calculate the remaining entries of column c */
   if(t->myid>0) {
   i=c+t->myid;
   ll=SPM_e(t->L,c,c);
   while (i<N) {
     temp=0;
     for(k=0; k<c; k++) {
      aa=SPM_e(t->L,i,k);
      bb=SPM_e(t->L,c,k);
      if((aa)&&(bb))
       temp+=aa*bb;
#ifdef DEBUG
      printf("[%d][%d] add [%d][%d]*[%d][%d]\n",i,c,i,k,c,k);
#endif
     }
     aa=(SPM_e(t->A,i,c)-temp)/ll;
     pthread_mutex_lock(t->lock_mutex);
      SPM_eq(t->L,i,c,aa);
     pthread_mutex_unlock(t->lock_mutex);
#ifdef DEBUG
     printf("thread %d, L[%d][%d]=%lf\n",t->myid,i,c,aa);
#endif
     i+=(*(t->T)-1);
    }
   }
   sync_barrier(t->barrier);
#ifdef DEBUG
   printf("col %d, leave id=%d\n",c,t->myid);
#endif
  }
  return NULL;
} 

/* Cholevsky decpmposition with thread support */
static int
cholevsky_decompose(SPMat *A, SPMat *L) {
 int M=3; 
 int i;
 tdata *thread_data;
 th_barrier b;
 pthread_t *th_array;
 pthread_attr_t attr;
 pthread_mutex_t lock_mutex;

  /* threads =3, one master, two slave */
/* thread related things */
 init_th_barrier(&b,M);
 pthread_attr_init(&attr);
 pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE);
 pthread_mutex_init(&lock_mutex,NULL);


 if ((th_array=(pthread_t*)malloc((size_t)M*sizeof(pthread_t)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((thread_data=(tdata*)malloc((size_t)M*sizeof(tdata)))==0) {
    fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
    exit(1);
  }

 for (i=0; i<M; i++) {
  thread_data[i].myid=i;
  thread_data[i].barrier=&b;
  thread_data[i].T=&M;
  thread_data[i].A=A;
  thread_data[i].L=L;
  thread_data[i].lock_mutex=&lock_mutex;
  pthread_create(&th_array[i],&attr,work_threadfn,(void*)(&thread_data[i]));
 } 

 for(i=0; i<M; i++) {
   pthread_join(th_array[i],NULL);
 } 

 destroy_th_barrier(&b);
 pthread_attr_destroy(&attr);
 pthread_mutex_destroy(&lock_mutex);
 free(th_array);
 free(thread_data);

 return 0;
}
#endif /* ! HAVE_PTHREAD */

#ifndef HAVE_PTHREAD
/* create cholevsky decomposition of sparse matrix A,
   output L, the sparse matrix.
  L should be initialized first
*/
static int
cholevsky_decompose(SPMat *A, SPMat *L)  {
  int i,j,k,N;
  double aa,bb,ll,temp; 
  N=SPM_dim(A);
  if(SPM_dim(L)!=N) return -1;
 
  for(j=0; j<N; j++) {
   temp=0;
   /* diagonal element of row j */
   for (k=0;k<j; k++) {
    aa=SPM_e(L,j,k);
    if(aa)
     temp+=aa*aa;
   }
   temp=SPM_e(A,j,j)-temp;

   if(temp<=0) {
    fprintf(stderr,"%s: %d: matrix not positive definite\n",__FILE__,__LINE__);
    temp=TOL;
   } else {
    temp=sqrt(temp);
   }
   SPM_eq(L,j,j,temp);
   ll=temp; /* remember this value */
   /* now find the j th column below the diagonal element */
   for (i=j+1; i<N; i++) {
    temp=0; 
    for(k=0; k<j; k++) {
     aa=SPM_e(L,i,k);
     bb=SPM_e(L,j,k);
     if((aa)&&(bb))
      temp+=aa*bb;
    }
    aa=(SPM_e(A,i,j)-temp)/ll;
    SPM_eq(L,i,j,aa);
   }
  }
  
  return 0;
}
#endif /* HAVE_PTHREAD */

/* AA - CSR format lower triangular matrix */
/* solve Ax = y using forward elimination */
static int
forward_eliminate(double *AA, int *JA, int *IA, int  N, int Nz, 
               double *x, double *y) {
  double temp,ll;
  int i,k,k1,k2;
  for(i=0; i<N; i++) {
   k1=IA[i];
   if (i<N-1) {
    k2=IA[i+1]-1;
   } else {
    k2=Nz-1;
   }
  /* range of row i is [k1,k2] */
  /* find the product AA(k1:k2) (except diagonal) with x(0:i-1) */
  /* note we have no ordering of row elements,
     so we do not know where the diagonal element is */
  ll=1;
  temp=0;
  for(k=k1; k<=k2; k++) {
   if(JA[k]!=i) { /* avoid diagonal element */
    temp+=AA[k]*x[JA[k]];
   } else {
    ll=AA[k];
   }
  }
  x[i]=(y[i]-temp)/ll;
 } 

 return 0;
}
/* solve Ax = y by back substitution, A is an upper triangular 
   matrix, given in CSR format */
static int
backward_substitute(double *AA, int *JA, int *IA, int N, int Nz,
      double *x, double *y)  {
  double temp,ll;
  int i,k,k1,k2;

  for (i=N-1; i>=0; i--) {
    k1=IA[i];
    if (i<N-1) {
     k2=IA[i+1]-1;
    } else {
     k2=Nz-1;
    }
   /* range for row i is [k1,k2]
   find the product of row i, AA(k1:k2) except diagonal
    with x[i+1:N-1] */
   temp=0;
   ll=1;
   for(k=k1;k<=k2; k++) {
    if(JA[k]!=i) {
     temp+=AA[k]*x[JA[k]];
    } else {
      ll=AA[k];
    }
   }
   x[i]=(y[i]-temp)/ll;
  }
  return 0;
}

/* find generalized eigenvalues of the pencil (A,B) using ARPACK
   store eigenvectors in E - size N x nev
   and eigenvalues in r - size nev x 1
   nev - the no. of eigenvalues to find
  lower_tiangular - if this is 1,
  NOTE: A, B are lower triangular, so need to fill up the upper triangle
  of the matrices. Otherwise, full matrices are assumed.
*/
int
arpack_find_eigs(SPMat *A, SPMat *B, MY_DOUBLE **E, MY_DOUBLE *r, MY_INT nev, int lower_triangular) {
 SPMat L,U;
 int N,i,j;

 double *AA,*AL, *AU;
 int *IA, *JA, *IL, *JL, *IU, *JU;
 unsigned int M,LL,UU;
 int ncv, lworkl;
 double *workd, *workl, *v, *resid;
 double *selection, *d;
 int info, ido, ierr;

 double *x, *y, temp;
 int maxiter;

 N=SPM_dim(A);

 if( lower_triangular ) {
 /* fill up the upper triangle of both matrices, first */
 for (i=0; i<N; i++) {
  for (j=0; j<i; j++) {
   temp=SPM_e(A,i,j);
   if(temp) 
    SPM_eq(A,j,i,temp);
   temp=SPM_e(B,i,j);
   if(temp) 
    SPM_eq(B,j,i,temp);
  }
 }
 }

 /* find cholevsky decomposition  of B*/
 SPM_init(&L,N);
 cholevsky_decompose(B, &L);
#ifdef DEBUG
 printf("A=[\n");
 for (i=0; i<N; i++) {
  for(j=0; j<N; j++) {
   printf("%lf ",SPM_e(A,i,j));
  }
  printf("\n");
 }
 printf("];\n");
#endif
#ifdef DEBUG
 printf("B=[\n");
 for (i=0; i<N; i++) {
  for(j=0; j<N; j++) {
   printf("%lf ",SPM_e(B,i,j));
  }
  printf("\n");
 }
 printf("];\n");
#endif


#ifdef DEBUG
 printf("L=[\n");
 for (i=0; i<N; i++) {
  for(j=0; j<N; j++) {
   printf("%lf ",SPM_e(&L,i,j));
  }
  printf("\n");
 }
 printf("];\n");
#endif
 printf("A: sparsity=%lf, B: sparsity=%lf, L: sparsity=%lf, size=%d\n",
     (double)SPM_size(A)/(double)(N*N),(double)SPM_size(B)/(double)(N*N),
     (double)SPM_size(&L)/(double)(N*N),N);
 M=SPM_size(A); 
 /* CSR arrays of A */
 if ((AA=(double *)calloc((size_t)M,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((JA=(int *)calloc((size_t)M,sizeof(int)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((IA=(int *)calloc((size_t)N,sizeof(int)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }

 SPM_build_lists(A);
 info=SPM_build_CSR(A,AA,JA,IA);
#ifdef DEBUG
 printf("A (%d) size=%d, CSR=%d\n",N,M,info);
#endif
 /* convert JA and IA to fortran format */
 for(i=0; i<N; i++) {
  JA[i]+=1;
  IA[i]+=1;
 }
 for(i=N; i<M; i++)
  JA[i]+=1;
#ifdef DEBUG1
 for(i=0; i<M; i++) {
  if(i<N) {
   printf("%lf : %d : %d\n",AA[i],JA[i],IA[i]);
  } else {
   printf("%lf : %d\n",AA[i],JA[i]);
  }
 }
#endif

 LL=SPM_size(&L);
 /* CSR arrays */
 if ((AL=(double *)calloc((size_t)LL,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((JL=(int *)calloc((size_t)LL,sizeof(int)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((IL=(int *)calloc((size_t)N,sizeof(int)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }

 SPM_build_lists(&L);
 info=SPM_build_CSR(&L,AL,JL,IL);
#ifdef DEBUG1
 printf("L size=%d, CSR=%d\n",LL,info);
 for(i=0; i<LL; i++) {
  if(i<N) {
   printf("%lf : %d : %d\n",AL[i],JL[i],IL[i]);
  } else {
   printf("%lf : %d\n",AL[i],JL[i]);
  }
 }
#endif

 /* create L^T as U */
 SPM_init(&U,N);
 SPM_transpose(&L,&U);
 SPM_destroy(&L);

 SPM_build_lists(&U);
 UU=SPM_size(&U);
 /* CSR arrays */
 if ((AU=(double *)calloc((size_t)UU,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((JU=(int *)calloc((size_t)UU,sizeof(int)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((IU=(int *)calloc((size_t)N,sizeof(int)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }

 info=SPM_build_CSR(&U,AU,JU,IU);

#ifdef DEBUG
 printf("L^T size=%d, CSR=%d\n",UU,info);
#endif
 SPM_destroy(&U);
#ifdef DEBUG1
 for(i=0; i<UU; i++) {
  if(i<N) {
   printf("%lf : %d : %d\n",AU[i],JU[i],IU[i]);
  } else {
   printf("%lf : %d\n",AU[i],JU[i]);
  }
 }
#endif

 /* convert arrays to fortran format */
 for(i=0; i<N; i++) {
  JU[i]+=1;
  JL[i]+=1;
  IU[i]+=1;
  IL[i]+=1;
 }
 for(i=N; i<UU; i++) {
  JU[i]+=1;
  JL[i]+=1;
 }

 maxiter=arpack_max_iterations;
 if(maxiter<0) maxiter=300;

 /* typically we set the subspace dimension to 4*nev,
   However, if we want a higher value, we use the global variable */
 ncv=4*nev;
 if(ncv <arpack_subspace_dimension) {
  ncv=arpack_subspace_dimension;
 } else {
  arpack_subspace_dimension=ncv;
 }
 if(ncv<4) ncv=4;
 if(ncv>N) ncv=N-1;

 lworkl=ncv*(ncv+8);
 /* ARPACK storage allocation */
 if ((workd=(double *)calloc((size_t)3*N,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((workl=(double *)calloc((size_t)lworkl,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((v=(double *)calloc((size_t)N*ncv,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((resid=(double *)calloc((size_t)N,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((selection=(double *)calloc((size_t)ncv,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((d=(double *)calloc((size_t)ncv*2,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }

 ardrive_(&N, &M, AA, JA, IA,
          &nev, &ncv, &lworkl, resid,
					v, workd, workl, &info, &ido,
					selection, d, &ierr,
          &LL, AL, JL, IL, &UU, AU, JU, IU, &maxiter);

 printf("info=%d, ido=%d, ierr=%d\n",info,ido,ierr);

 free(AA);
 free(JA);
 free(IA);
 free(AL);
 free(JL);
 free(IL);
 free(workd);
 free(workl);
 free(resid);
 free(selection);

 /* convert arrays back to C format */
 for(i=0; i<N; i++) {
  JU[i]-=1;
  IU[i]-=1;
 }
 for(i=N; i<UU; i++) {
  JU[i]-=1;
 }
 if ((x=(double *)calloc((size_t)N,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((y=(double *)calloc((size_t)N,sizeof(double)))==0) {
   fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 for(i=0; i<nev; i++){
    r[i]=d[i];
    for(j=0; j<N; j++) 
     y[j]=v[i*N+j];
    backward_substitute(AU, JU, IU, N, UU, x, y);
    for(j=0; j<N; j++)
     E[j][i]=x[j];
 }

 free(d);
 free(v);
 free(AU);
 free(IU);
 free(JU);


 free(x);
 free(y);
 return 0;
}

/* dummy drivers when we do not use ARPACK */
#ifndef USE_ARPACK
void dsaupd_(int *ido, char *bmat, int *n, char *which,
                int *nev, double *tol, double *resid, int *ncv,
                double *v, int *ldv, int *param, int *ipntr,
                double *workd, double *workl, int *lworkl,
                int *info, unsigned long bmat_len, unsigned long which_len) {
  *ido=0;
  *info=0;
}

void dseupd_(int *rvec, char *all, int *select_, double *d,
                double *v1, int *ldv1, double *sigma,
                char *bmat, int *n, char *which, int *nev,
                double *tol, double *resid, int *ncv, double *v,
                int *ldv, int *iparam, int *ipntr, double *workd,
                double *workl, int *lworkl, int *ierr) {
 *ierr=0;
}
#endif /* USE_ARPACK */
