/*  pdnmesh - a 2D finite element solver
    Copyright (C) 2001-2005 Sarod Yatawatta <sarod@users.sf.net>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  $Id: spmat.c,v 1.13 2005/04/21 23:16:55 sarod Exp $
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef HAVE_PTHREAD
#include <pthread.h>
#endif
#include "types.h"

extern RBT_node_type sentinel;
#define NIL &sentinel


/* data record type */
typedef struct record_{
    int key; /* column */
    MY_DOUBLE z; /* value */
} record_type;

static int
equ(const void*a,const void *b)
{ 
  record_type *p,*q;
  p=(record_type *)a;
  q=(record_type *)b;
  return(p->key==q->key);
}
static int
lt(const void*a,const void *b)
{ 
  record_type *p,*q;
  p=(record_type *)a;
  q=(record_type *)b;
  return(p->key<q->key);

}

static void
print_details(const void *a)
{
  record_type *p;
  p=(record_type *)a;
  printf("(%d)%lf\t",p->key,p->z);
}

static void
delete_record(void *a)
{
 free((record_type*)a);
}

/* list data record type = point to record_type  */
typedef struct l_record_{
  record_type *l;
} list_record_type;


/* delete list element */
static void
delete_list_record(void *a) {
 /* record deletion is handled by the tree, so we do not 
  do anything here  except free the pointer */
 free((list_record_type*)a);
}

/* intialize a matrix */
int
SPM_init( SPMat *sM, int N ) {
  int i;
  sM->N=N;
  sM->T=0;
  /* allocate N trees */
	if ((sM->rt=(RBT_tree**)malloc(sizeof(RBT_tree*)*(size_t)N))==0) {
       fprintf(stderr,"%s: %d: no free memory",__FILE__,__LINE__);
       exit(1);
 }

 for (i=0; i<N; i++) {
	if ((sM->rt[i]=(RBT_tree*)malloc(sizeof(RBT_tree)))==0) {
       fprintf(stderr,"%s: %d: no free memory",__FILE__,__LINE__);
       exit(1);
 }
 }
/* initialize trees */
 for (i=0; i<N; i++) {
  RBT_init(sM->rt[i],&equ,&lt,&print_details,&delete_record);
 }
 
 /* allocate N lists, initialize them */
 if ((sM->list=(glist**)malloc(sizeof(glist*)*(size_t)N))==0) {
       fprintf(stderr,"%s: %d: no free memory",__FILE__,__LINE__);
       exit(1);
 }

 for (i=0; i<N; i++) {
	if ((sM->list[i]=(glist*)malloc(sizeof(glist)))==0) {
       fprintf(stderr,"%s: %d: no free memory",__FILE__,__LINE__);
       exit(1);
 }
 }
/* initialize lists */
 for (i=0; i<N; i++) {
  glist_init(sM->list[i],sizeof(list_record_type*),&delete_list_record);
 }
 return(0);
}

/* add z to i,j location of M */
/* initially assumed zero */
/* i,j has to be within [0,N-1] */
int
SPM_add( SPMat *sM, int i, int j,  MY_DOUBLE z) {
  record_type *rec,brec; 
  int status;
if (i<0 || i>= sM->N|| j<0 || j>= sM->N) return(-1); /* error in range */
brec.key = j;
rec=RBT_find(&brec, sM->rt[i]);
if (rec) {
      rec->z+=z;
      /* if record is zero do deletion */
     if (IS_ZERO(rec->z)) {
      RBT_delete((void*)rec,sM->rt[i]);
	    free(rec);
      sM->T--; /* decrease element count */
     }
} else {
    	if ((rec=(record_type *)malloc(sizeof(*rec)))==0) {
       fprintf(stderr,"%s: %d: no free memory",__FILE__,__LINE__);
       exit(1);
     }
     rec->key = brec.key;
     rec->z= z;
     status=RBT_insert(rec, sM->rt[i]);
     if (status==RBT_STATUS_OK) {
      sM->T++; /* increase element count */
				 } else if (status==RBT_STATUS_KEY_FOUND) {
        free(rec);
				 } else {
      free(rec);
		   }
}

return(0);
}



/* myltiply i,j location of M  by z */
/* initially assumed zero */
/* i,j has to be within [0,N-1] */
int
SPM_mult( SPMat *sM, int i, int j,  MY_DOUBLE z) {
  record_type *rec,brec; 
if (i<0 || i>= sM->N|| j<0 || j>= sM->N) return(-1); /* error in range */
brec.key = j;
brec.z=z;
rec=RBT_find(&brec, sM->rt[i]);
if (rec) {
      rec->z*=z;
      /* if record is zero do deletion */
     if (IS_ZERO(rec->z)) {
      RBT_delete((void*)rec,sM->rt[i]);
	    free(rec);
      sM->T--; /* decrease element count */
     }
} /* if rec not found, no need to multiply because
  result will be zero anyway */ 
return(0);
}


/* returns dimension of Matrix, or value of N */
int
SPM_dim(SPMat *sM) {
 return(sM->N);
}

/* returns filled size of M, or value of N */
unsigned long int
SPM_size(SPMat *sM) {
 return(sM->T);
}
/* delete matrix M */
int
SPM_destroy(SPMat *sM) {
 int i;
/* free all trees */
 for (i=0; i<sM->N; i++) {
  RBT_free(sM->rt[i]);
  free(sM->rt[i]);
 }
 free(sM->rt);
 /* free all lists */
 for (i=0; i<sM->N; i++) {
  glist_delete(sM->list[i]);
  free(sM->list[i]);
 }
 free(sM->list);

return(0);
}


static  MY_DOUBLE
visit_list_product(glist *L,MY_DOUBLE *x) {
 MY_DOUBLE temp; 
 /* assume x has valid range */
 list_record_type *p; 
 /* initialize list iterator */
 glist_set_iter_forward(L);
 /* iterate list */
 p=(list_record_type*)glist_iter_forward(L);
 temp=0;
 while(p) {
  if (p->l) {
#ifdef DEBUG
   printf("%d: %lf\n",p->l->key,p->l->z);
#endif
   temp+=x[p->l->key]*(p->l->z);
  }
  p=(list_record_type*)glist_iter_forward(L);
 }
 return temp;
}


/* Matrix vector product A.x - using linked list */
/* A-sparse matrix*/
/* x- vector*/
/* v- result = Ax */
#ifndef HAVE_PTHREAD
int 
SPM_vec(SPMat *sM, MY_DOUBLE *x, MY_DOUBLE *v) {
 MY_INT i;
 for (i=0; i<sM->N; i++) {
 v[i]=visit_list_product(sM->list[i],x);
 }
 return(0);
}
#endif /* HAVE_PTHREAD */
#ifdef HAVE_PTHREAD
typedef struct thread_data_ {
 int myid; /* thread id */
 SPMat *sM; /* matrix */
 MY_DOUBLE *v; /* result */
 MY_DOUBLE *x; /* input vector */
 int T; /* total no. of threads */
} thread_data_t;

/* thread function */
static void *
work_threadfn(void *data) {
 thread_data_t *t=(thread_data_t*)data;
 int m;
 if(t==0) return NULL;
 m=t->myid;
 while(m<t->sM->N) {
 /* since each thread operates on independent rows, 
  no need for locking */
  t->v[m]=visit_list_product(t->sM->list[m],t->x);
  m=m+t->T; /* increment row by thread count */
 }
#ifdef DEBUG
 printf("thread %d finished\n",t->myid);
#endif
 return NULL;
}

int 
SPM_vec(SPMat *sM, MY_DOUBLE *x, MY_DOUBLE *v) {
 MY_INT i;
 int M=2; /* two threads */
 pthread_attr_t attr;
 pthread_t *th_array;
 thread_data_t *thread_data;
  
 pthread_attr_init(&attr);
 pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE);

 if ((th_array=(pthread_t*)malloc((size_t)M*sizeof(pthread_t)))==0) {
   fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
   exit(1);
 }
 if ((thread_data=(thread_data_t*)malloc((size_t)M*sizeof(thread_data_t)))==0) {
    fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
    exit(1);
 }
 for (i=0; i<M; i++) {
  thread_data[i].myid=i;
  thread_data[i].T=M;
  thread_data[i].sM=sM;
  thread_data[i].x=x;
  thread_data[i].v=v;

  pthread_create(&th_array[i],&attr,work_threadfn,(void*)(&thread_data[i]));
 }
 for(i=0; i<M; i++) {
   pthread_join(th_array[i],NULL);
 }

 pthread_attr_destroy(&attr);
 free(th_array);
 free(thread_data);

 return(0);
}
#endif /* !HAVE_PTHREAD */

/* local visit */
static void 
visit_node_copy(RBT_node_type *n,RBT_tree *t, MY_DOUBLE *v)
{
   record_type *p;
  if ( n != NIL ){
    visit_node_copy(n->left,t,v);
    p=(record_type *)n->rec;
    v[p->key]=p->z;
    visit_node_copy(n->right,t,v);
  }
}

/* local visit-product */
static MY_DOUBLE 
visit_node_product(RBT_node_type *n,RBT_tree *t, MY_DOUBLE *v)
{
   record_type *p;
  if ( n != NIL ){
    p=(record_type *)n->rec;
    return(v[p->key]*(p->z)+visit_node_product(n->left,t,v)+visit_node_product(n->right,t,v));
  } else {
   return(0);
  }
}

/* traverse tree i and */
/* copy row i of M to vector v */
/* v size N by 1 */
int 
SPM_row(SPMat *sM, MY_INT i, MY_DOUBLE *v) {
 if ( i<0 || i>=sM->N) return(-1); /* error range */
 visit_node_copy(sM->rt[i]->root,sM->rt[i],v);
 return(0);
}


/* Matrix vector product A.x*/
/* A-sparse matrix*/
/* x- vector*/
/* v- result = Ax */
int 
SPM_vec_recursive(SPMat *sM, MY_DOUBLE *x, MY_DOUBLE *v) {
 MY_INT i;
 for (i=0; i<sM->N; i++) {
 v[i]=visit_node_product(sM->rt[i]->root,sM->rt[i],x);
 }
 return(0);
}

/* returns the value at location (i,j)
 of the matrix or zero if none found */
MY_DOUBLE 
SPM_e(SPMat *sM, MY_INT i, MY_INT j) {
  record_type *rec,brec; 
if (i<0 || i>= sM->N|| j<0 || j>= sM->N) return(0); /* error in range */
brec.key = j;
rec=RBT_find(&brec, sM->rt[i]);
if (rec)
 return(rec->z);
/* else */
return(0);
}

/* copy column i of M to vector v */
/* v size N by 1 */
/* note column access is more expensive than row access */
int 
SPM_col(SPMat *sM, MY_INT i, MY_DOUBLE *v) {
 int k;
 if ( i<0 || i>=sM->N) return(-1); /* error range */
 for (k=0; k<sM->N; k++) {
  v[k]=SPM_e(sM,k,i);
 }
 return(0);
}


/* replaces the value at location (i,j)
 of the matrix with z */
int
SPM_eq(SPMat *sM, MY_INT i, MY_INT j,MY_DOUBLE z) {
  record_type *rec,brec; 
if (i<0 || i>= sM->N|| j<0 || j>= sM->N) return(0); /* error in range */
brec.key = j;
rec=RBT_find(&brec, sM->rt[i]);
if (rec) {
 rec->z=z;
    /* if record is zero do deletion */
    if (IS_ZERO(rec->z)) {
     RBT_delete((void*)rec,sM->rt[i]);
     free(rec);
    sM->T--;
    }
 return(0);
}
/* else */
 if (!IS_ZERO(z))
  SPM_add( sM, i, j,  z);
return(-1);
}


/* local visit */
static void 
visit_node_copy_mat(RBT_node_type *n,RBT_tree *t, SPMat *sM, MY_INT row)
{
  record_type *p;
  if ( n != NIL ){
    visit_node_copy_mat(n->left,t,sM,row);
    p=(record_type *)n->rec;
    SPM_eq(sM, row, p->key, p->z);
    visit_node_copy_mat(n->right,t,sM,row);
  }
}

/* copy matrix sM to matrix B */
int 
SPM_copy(SPMat *sM, SPMat *B) {
 int i;
 int N=SPM_dim(sM);
 if(N==SPM_dim(B)) { 
  for(i=0; i<N; i++) 
   visit_node_copy_mat(sM->rt[i]->root,sM->rt[i],B,i);
  return(0);
  B->T=sM->T;
 }
 /* else */
 return(-1);
}

/* local visit */
static void 
visit_node_copy_mat_scalar(RBT_node_type *n,RBT_tree *t, SPMat *sM, MY_INT row, MY_DOUBLE c)
{
  record_type *p;
  if ( n != NIL ){
    visit_node_copy_mat_scalar(n->left,t,sM,row,c);
    p=(record_type *)n->rec;
    SPM_add(sM, row, p->key, c*(p->z));
    visit_node_copy_mat_scalar(n->right,t,sM,row,c);
  }
}

/* add matrix sM multiplied by scalar c  to matrix B */
/* B = B+ sM x c */
int 
SPM_add_scalar(SPMat *sM, SPMat *B, MY_DOUBLE c) {
 int i;
 int N=SPM_dim(sM);
 if(N==SPM_dim(B)) { 
  for(i=0; i<N; i++) 
   visit_node_copy_mat_scalar(sM->rt[i]->root,sM->rt[i],B,i,c);
  return(0);
 }
 /* else */
 return(-1);
}

/* local visit  - insert to list */
static void 
visit_node_insert_list(RBT_node_type *n,RBT_tree *t, glist *List)
{
  list_record_type p;
  if ( n != NIL ){
    visit_node_insert_list(n->left,t,List);
    p.l=(record_type *)n->rec;
    /* insert list element */
    glist_insert(List,(void*)&p);
#ifdef DEBUG
    printf("%d: %lf\n",p.l->key,p.l->z);
#endif
    visit_node_insert_list(n->right,t,List);
  }
}

/* build linked lists of each row elements for serial access */
/* returns 0 if no error */
int
SPM_build_lists(SPMat *sM) {
 int i;
 int N=SPM_dim(sM);
 for (i=0; i<N; i++) {
  /* free list if list is not empty */
  if(!glist_empty(sM->list[i])) 
   /* if not empty, destroy the list first */
   glist_delete(sM->list[i]);
  /* now the list is empty */
  visit_node_insert_list(sM->rt[i]->root,sM->rt[i],sM->list[i]);  
 }
 return 0;
}
/* traverse the list and add the elements
  to the arrays AA and AJ
  AA[k] will store the non zero value
  AJ[k] will store the column number of the non zero value
  returns the number of non-zero values stored
*/
static int
visit_list_add_array(glist *L,MY_DOUBLE *AA, MY_INT *AJ) {
 MY_INT t; 
 /* assume x has valid range */
 list_record_type *p; 
 /* initialize list iterator */
 glist_set_iter_forward(L);
 /* iterate list */
 p=(list_record_type*)glist_iter_forward(L);
 t=0;
 while(p) {
  if (p->l) {
#ifdef DEBUG
   printf("%d: %lf\n",p->l->key,p->l->z);
#endif
   AA[t]=p->l->z;
   AJ[t]=p->l->key;
   t++;
  }
  p=(list_record_type*)glist_iter_forward(L);
 }
 return t;
}

/* build compressed sparse row (CSR) storage format arrays
  from the given sparse matrix.
  AA - size Nz x 1 : array of non zero elements
  AJ - size Nz x 1: array  of column indices
  AI - size N x 1: array of indices for starting point of rows
  returns the final filled position of array AA
*/
int
SPM_build_CSR(SPMat *A, double *AA, int* JA, int*IA) {
 int offset,i,k,N;
 /* assume all arrays have the required storage allocated */
 offset=0;
 N=SPM_dim(A);
 for (i=0; i<N; i++) {
  k=visit_list_add_array(A->list[i],&AA[offset],&JA[offset]);
  /* if row is empty, make IA[i]=-1 */
  if(!k) {
   fprintf(stderr,"%s: %d: row missing from matrix\n",__FILE__,__LINE__);
  }
  IA[i]=offset;
  offset+=k;
 }
 return offset;
}


/* create a transpose of a matrix
   by visiting the lists */
/* B = A^T */
int
SPM_transpose(SPMat *A, SPMat *B) {
  list_record_type *p;
  int i,N;
  N=SPM_dim(A);
  if(SPM_dim(B)!=N) return -1;

  for(i=0; i<N; i++) {
   glist_set_iter_forward(A->list[i]);
   p=(list_record_type*)glist_iter_forward(A->list[i]);
   while(p) {
    if(p->l) {
      SPM_eq(B,p->l->key,i,p->l->z);
    }
    p=(list_record_type*)glist_iter_forward(A->list[i]);
   }
  }
  return 0;
}
